/**
 * The main function.
 *
 * (Executed at the bottom)
 */
function main() {

    // Apply icxml only when a file path is provided as an arguemnt to the script
    // Reads the file path from the argument  
    var icxmlPath = app.scriptArgs.getValue("icxmlPath");
    // var icxmlPath = "C:/Users/UTGO/Desktop/icxml_for_novo_template.xml";
    if (!icxmlPath) {
        return;
    }

    // Apply the icxml contents on all  the xml elements of the document
    applyICXMLContentOnDocument(icxmlPath);
}

/**
 * Apply the content from icxml file on the available XML tags in the document.
 * @param {*} icxmlPath 
 */
function applyICXMLContentOnDocument(icxmlPath) {
    // Read the icxml file
    var xmlFile = File(icxmlPath);
    xmlFile.open("r");

    // Create an XML object from icxml file contents
    var icxml = XML(xmlFile.read());
    xmlFile.close();

    // Get the top XMLItems in the active document.
    var xmlElements = app.activeDocument.xmlElements;

    // Get all nested XMLElements in those.
    var allXmlElements = getAllXmlElements(xmlElements);

    // Get the content.
    var content = getContent(icxml);

    // Apply the content to the XML Elements.
    applyAllContent(content, allXmlElements);
}

/**
 * Parse the icxml and return the content to import.
 */
function getContent(icxml) {

    // Root of the xml that contains all artwork content
    var artworkContent = icxml.child("ArtworkContentCopyElements");

    // List of all the parsed content objets
    var tagContentList = [];

    var inddContentElements = artworkContent[0].child(
        "ArtworkContentCopyElement"
    );

    // Iterate over the array of contents having "Tag" and "TextContent"
    for (var i = 0; i < inddContentElements.length(); i++) {
        var inddContentElement = inddContentElements[i];
        var iddTag = inddContentElement.child("Tag")[0];
        var iddContent = inddContentElement
            .child("TextContent")[0]
            .child("IddContent")[0]
            .text();

        // Create an Object for each content and add them into the list
        var tagContent = {};
        tagContent[iddTag] = iddContent;
        tagContentList.push(tagContent);
    }

    return tagContentList;
}

/**
 * Return all XMLElemnets that are nested in the passed xml items.
 *
 * @param {*} xmlElements
 */
function getAllXmlElements(xmlElements) {
    // The tagged elements in the stories
    var allXmlElements = [];

    // How many elements were passed?
    var nXmlElements = xmlElements.count();
    for (var i = 0; i < nXmlElements; i++) {
        // The XML item
        var xmlElement = xmlElements.item(i);

        // Add it to the nested xml elements.
        allXmlElements.push(xmlElement);

        // Get the XML items directly in this element.
        var childXmlElements = xmlElement.xmlElements;
        if (childXmlElements.count() > 0) {
            // Get the XML elements nested in those.
            var childAllXmlElements = getAllXmlElements(childXmlElements);

            // If there were found any
            if (childAllXmlElements.length > 0) {
                // Add them to the list.
                allXmlElements = allXmlElements.concat(childAllXmlElements);
            }
        }
    }
    return allXmlElements;
}

/**
 * Apply the content.
 *
 * @param {Object} content
 * @param {Array XMLElement} xmlElements
 */
function applyAllContent(content, xmlElements) {
    // Iterate over the content.
    for (var item = 0; item < content.length; item++) {
        var iddContentObj = content[item];

        // Go over the XMLElements that match this tagName.
        for (var i = 0; i < xmlElements.length; i++) {
            // The element and its tagname.
            var xmlElement = xmlElements[i];
            var markupTagName = xmlElement.markupTag.name;

            // Check if it matches.
            if (iddContentObj.hasOwnProperty(markupTagName)) {
                var iddContent = iddContentObj[markupTagName].toString();
                // Apply the content to this tag.
                applyContent(iddContent, xmlElement);
            }
        }
    }
}

/**
 * Apply the content string to the element.
 *
 * @param {String} string
 * @param {XMLElement} xmlElement
 */
function applyContent(string, xmlElement) {
    // Detect the "default" paragraph style.
    var initParagraphStyle = getInitParagraphStyle(xmlElement);

    // Get the "No" character style.
    var characterStyles = app.activeDocument.characterStyles;
    var noCharacterStyle = undefined;
    for (var i = 0; i < characterStyles.count(); i++) {
        if (characterStyles.item(i).name == "[None]") {
            noCharacterStyle = characterStyles.item(i);
        }
    }

    // Clear the contents of the XML element.
    xmlElement.contents = "";

    // Try to parse the string.
    XML.ignoreWhitespace = false;
    var bodyElement = new XML(string);

    // Assume that this is a "<body>" tag.
    if (
        bodyElement.nodeKind() != "element" ||
        bodyElement.localName() != "body"
    ) {
        throw "Encountered invalid element!";
    }

    // Iterate over the paragraphs in this body element.
    for (var i = 0; i < bodyElement.children().length(); i++) {
        // The paragraph element.
        var paragraphElement = bodyElement.child(i);

        // Make sure that we have a paragraph element.
        if (
            paragraphElement.nodeKind() != "element" ||
            paragraphElement.localName() != "p"
        ) {
            // Skip over it.
            continue;
        }

        // Start with a newline if this is not the first paragraph
        if (xmlElement.paragraphs.count() != 0 && xmlElement.contents.length > 0) {
            xmlElement.contents += "\r"; // Don't use the classic '\n' as this will not create a new paragraph.
        }

        // Iterate over the text element nodes in this paragraph.
        for (var j = 0; j < paragraphElement.children().length(); j++) {
            // The node in the paragraph.
            var node = paragraphElement.child(j);

            // Check whether we're working with unstyled paragraph text ...
            if (node.nodeKind() == "text") {
                // Remember where we start
                var styleStartIndex = 0;
                if (xmlElement.characters.length > 0) {
                    styleStartIndex = xmlElement.characters.length;
                }

                // Check for atleast single character
                var nodeString = node.toString();
                if (nodeString.length <= 0) {
                    continue;
                }

                // Add the contents.
                xmlElement.contents += nodeString;

                // Don't apply character style on only whitespace
                if (nodeString.match("^\\s+$")) {
                    continue;
                }

                // Remember where we ended.
                var styleEndIndex = styleStartIndex + nodeString.length - 1;

                // Make sure that no character style is applied.
                // Get the range.
                var styleRange = xmlElement.characters.itemByRange(
                    styleStartIndex,
                    styleEndIndex
                );

                // Make sure that no character style is applied.
                styleRange.applyCharacterStyle(noCharacterStyle);
            }
            // Or whether a
            else if (node.nodeKind() == "element" && node.localName() == "span") {
                // Remember where we start
                var styleStartIndex = 0;
                if (xmlElement.characters.length > 0) {
                    styleStartIndex = xmlElement.characters.length;
                }
                // Check for atleast single character
                var nodeString = node.toString();
                if (nodeString.length <= 0) {
                    continue;
                }

                // Add the contents.
                xmlElement.contents += nodeString;

                // Don't apply character style on only whitespace
                if (nodeString.match("^\\s+$")) {
                    continue;
                }

                // Remember where we ended.
                var styleEndIndex = styleStartIndex + nodeString.length - 1;

                // Get the range.
                var styleRange = xmlElement.characters.itemByRange(
                    styleStartIndex,
                    styleEndIndex
                );

                // Extract a possibly hinted style
                var suggestedCharacterStyleString = node.attribute("cstyle").toString();

                // Deduce the character style
                characterStyle = findCharacterStyleFor(
                    suggestedCharacterStyleString,
                    initParagraphStyle
                );

                // Apply the character style
                if (characterStyle) {
                    styleRange.applyCharacterStyle(characterStyle);
                }
            }
            // Or something unexpected ...
            else {
                throw "Encountered invalid element!";
            }
        }

        // Extract a possibly hinted style
        var suggestedParagraphStyleString = paragraphElement
            .attribute("pstyle")
            .toString();

        // Deduce the paragraph style.
        var paragraphStyle = findParagraphStyleFor(
            suggestedParagraphStyleString,
            initParagraphStyle
        );

        // Apply the appropiate paragraph style.
        xmlElement.paragraphs.lastItem().applyParagraphStyle(paragraphStyle);
    }
}

/**
 * Get the intial paragraph style for a particular xml element.
 * @param {XMLElement} xmlElement
 */
function getInitParagraphStyle(xmlElement) {
    // Find the "Basic" and "No" paragraph styles.
    var paragraphStyles = app.activeDocument.paragraphStyles;
    var basicParagraphStyle = undefined;
    var noParagraphStyle = undefined;
    for (var i = 0; i < paragraphStyles.count(); i++) {
        if (paragraphStyles.item(i).name == "[Basic Paragraph]") {
            basicParagraphStyle = paragraphStyles.item(i);
        } else if (paragraphStyles.item(i).name == "[No Paragraph Style]") {
            noParagraphStyle = paragraphStyles.item(i);
        }
    }

    // Check if there's a paragraph already in the element.
    if (xmlElement.paragraphs.count() > 0) {
        // Get the first paragraph.
        var initFirstParagraph = xmlElement.paragraphs.firstItem();

        // Return the style of the first paragraph.
        var paragraphStyle = initFirstParagraph.appliedParagraphStyle;

        // Check if it's not the "Basic" or "No" paragraph styles
        if (
            paragraphStyle.name != basicParagraphStyle.name &&
            paragraphStyle.name != noParagraphStyle.name
        ) {
            return paragraphStyle;
        }
    }
    // No paragraph in the element or it hase the "Basic" or "No" paragraph styles.

    // Create a new "Default" paragraph style based on the basic paragraph style.
    var paragraphStyleName = "Default";
    return getParagraphStyle(paragraphStyleName, basicParagraphStyle);
}

/**
 * Find a suitable character style.
 *
 * @param {*} suggestedStyle The style suggested by the style span.
 * @param {*} initParagraphStyle The initial paragraph style of the XML element.
 */
function findCharacterStyleFor(suggestedStyleString, initParagraphStyle) {
    // See if a style was suggested.
    if (
        suggestedStyleString == undefined ||
        suggestedStyleString == null ||
        suggestedStyleString.length == 0
    ) {
        // No style was suggested, return the initial one.
        return undefined;
    }
    // A style was suggested
    else {
        // Analyze the name of the initial paragraph style and try to get a prefix
        var prefix = "";
        var initParagraphStyleName = initParagraphStyle.name;
        var initParagraphStyleNameParts = initParagraphStyleName.split(" ");
        if (initParagraphStyleNameParts.length > 0) {
            prefix = initParagraphStyleNameParts[0];
        }

        // The character style name.
        var characterStyleName = prefix + " " + suggestedStyleString;

        // Return that character style (create if necessary)
        return getCharacterStyle(characterStyleName);
    }
}

/**
 * Find a suitable paragraph style.
 *
 * @param {*} suggestedStyle The style suggested by the paragraph.
 * @param {*} initParagraphStyle The initial paragraph style of the XML element.
 */
function findParagraphStyleFor(suggestedStyleString, initParagraphStyle) {
    // See if a style was suggested.
    if (
        suggestedStyleString == undefined ||
        suggestedStyleString == null ||
        suggestedStyleString.length == 0
    ) {
        // No style was suggested, return the initial one.
        return initParagraphStyle;
    }
    // A style was suggested
    else {
        // Analyze the name of the initial paragraph style and try to get a prefix
        var prefix = "";
        var initParagraphStyleName = initParagraphStyle.name;
        var initParagraphStyleNameParts = initParagraphStyleName.split(" ");
        if (initParagraphStyleNameParts.length > 0) {
            prefix = initParagraphStyleNameParts[0];
        }

        // The paragraph style name.
        var paragraphStyleName = prefix + " " + suggestedStyleString;

        // Get that paragraph style (create if necessary)
        return getParagraphStyle(paragraphStyleName, initParagraphStyle);
    }
}

/**
 * Get a character style and create it if it doesn't exist.
 *
 * @param {*} characterStyleName
 */
function getCharacterStyle(characterStyleName) {
    // Check if such a style name does exist.
    var characterStyle;
    var characterStyles = app.activeDocument.allCharacterStyles;
    for (var i = 0; i < characterStyles.length; i++) {
        if (characterStyles[i].name == characterStyleName) {
            characterStyle = characterStyles[i];
            break;
        }
    }

    // If no such character style was found, create one.
    if (characterStyle == undefined) {
        // Create a new character style
        characterStyle = app.activeDocument.characterStyles.add({
            name: characterStyleName
        });

        // Make some guesses on how the character style should look based on the name.
        baseCharacterStylingOnName(characterStyle);

        // Move it to the MISSING character style group.
        var characterStyleGroup;
        var characterStyleGroups = app.activeDocument.characterStyleGroups;
        for (var i = 0; i < characterStyleGroups.length; i++) {
            if (characterStyleGroups.item(i).name == "MISSING") {
                characterStyleGroup = characterStyleGroups.item(i);
                break;
            }
        }

        // If it does not exist, create it.
        if (characterStyleGroup == undefined) {
            characterStyleGroup = characterStyleGroups.add({
                id: "MISSING",
                name: "MISSING"
            });
        }

        // Move the style to the MISSING group (if not already contained)
        var inMissing = false;
        for (var i = 0; i < characterStyleGroup.characterStyles.count(); i++) {
            if (
                characterStyleGroup.characterStyles.item(i).name == characterStyle.name
            ) {
                inMissing = true;
                break;
            }
        }
        if (!inMissing) {
            characterStyle.move(LocationOptions.AT_END, characterStyleGroup);
        }
    }
    return characterStyle;
}

/**
 * Get a paragraph style and create it if it doesn't exist.
 *
 * @param {*} paragraphStyleName
 */
function getParagraphStyle(paragraphStyleName, referenceParagraphStyle) {
    // Check if such a style name does exist.
    var paragraphStyle;
    var paragraphStyles = app.activeDocument.allParagraphStyles;
    for (var i = 0; i < paragraphStyles.length; i++) {
        if (paragraphStyles[i].name == paragraphStyleName) {
            paragraphStyle = paragraphStyles[i];
            break;
        }
    }

    // If no such paragraph style was found, create one.
    if (paragraphStyle == undefined) {
        // Duplicate the style from the initial one.
        paragraphStyle = referenceParagraphStyle.duplicate();
        paragraphStyle.name = paragraphStyleName;

        // Make some guesses on how the paragraph style should look based on the name.
        baseParagraphStylingOnName(paragraphStyle);

        // Move it to the MISSING paragraph style group.
        var paragraphStyleGroup;
        var paragraphStyleGroups = app.activeDocument.paragraphStyleGroups;
        for (var i = 0; i < paragraphStyleGroups.length; i++) {
            if (paragraphStyleGroups.item(i).name == "MISSING") {
                paragraphStyleGroup = paragraphStyleGroups.item(i);
                break;
            }
        }

        // If it does not exist, create it.
        if (paragraphStyleGroup == undefined) {
            paragraphStyleGroup = paragraphStyleGroups.add({
                id: "MISSING",
                name: "MISSING"
            });
        }

        // Move the style to the MISSING group (if not already contained)
        var inMissing = false;
        for (var i = 0; i < paragraphStyleGroup.paragraphStyles.count(); i++) {
            if (
                paragraphStyleGroup.paragraphStyles.item(i).name == paragraphStyle.name
            ) {
                inMissing = true;
                break;
            }
        }
        if (!inMissing) {
            paragraphStyle.move(LocationOptions.AT_END, paragraphStyleGroup);
        }
    }
    return paragraphStyle;
}

/**
 * Make some decision on how this character style should look based on the name.
 *
 * @param {CharacterStyle} characterStyle
 */
function baseCharacterStylingOnName(characterStyle) {
    // Get the name of the style.
    var name = characterStyle.name.toLowerCase();

    // Guess Bold
    var fontStyle = "";
    if (name.indexOf("bold") >= 0) {
        fontStyle += "Bold";
    }
    // Guess Italic
    if (name.indexOf("italic") >= 0) {
        // Add  a space if not empty.
        if (fontStyle.length > 0) {
            fontStyle += " ";
        }
        fontStyle += "Italic";
    }
    // Apply the font style.
    if (fontStyle.length > 0) {
        characterStyle.fontStyle = fontStyle;
    }

    // Guess Subscript
    if (name.indexOf("sub") >= 0) {
        characterStyle.position = Position.SUBSCRIPT;
    }

    // Guess Superscript
    if (name.indexOf("super") >= 0) {
        characterStyle.position = Position.SUPERSCRIPT;
    }

    // Guess underline
    if (name.indexOf("underline") >= 0) {
        characterStyle.underline = true;
    }

    // Guess strike
    if (name.indexOf("strike") >= 0) {
        characterStyle.strikeThru = true;
    }
}

/**
 * Make some decision on how this paragraph style should look based on the name.
 *
 * @param {ParagraphStyle} paragraphStyle
 */
function baseParagraphStylingOnName(paragraphStyle) {
    // Get the name of the style.
    var name = paragraphStyle.name.toLowerCase();

    // If this is an unordered item.
    if (name.indexOf("unordered item") >= 0) {
        paragraphStyle.bulletsAndNumberingListType = ListType.BULLET_LIST;
        paragraphStyle.bulletsTextAfter = "\t";
        paragraphStyle.leftIndent = "1p0";
        paragraphStyle.firstLineIndent = "-1p0";
    }
    // If this is an ordered item.
    else if (name.indexOf("ordered item") >= 0) {
        paragraphStyle.bulletsAndNumberingListType = ListType.NUMBERED_LIST;
        paragraphStyle.bulletsTextAfter = "\t";
        paragraphStyle.leftIndent = "1p0";
        paragraphStyle.firstLineIndent = "-1p0";
    }
    // If this is an unordered item.
    else if (name.indexOf("unordered subitem") >= 0) {
        paragraphStyle.bulletsAndNumberingListType = ListType.BULLET_LIST;
        paragraphStyle.bulletsTextAfter = "\t";
        paragraphStyle.leftIndent = "2p0";
        paragraphStyle.firstLineIndent = "-2p0";
    }
    // If this is an ordered item.
    else if (name.indexOf("ordered subitem") >= 0) {
        paragraphStyle.bulletsAndNumberingListType = ListType.NUMBERED_LIST;
        paragraphStyle.bulletsTextAfter = "\t";
        paragraphStyle.leftIndent = "2p0";
        paragraphStyle.firstLineIndent = "-2p0";
    }
}



// Execute the main method
main();